""" fprime_gds.test_plugins:

A set of pytest-enabled tests that are used to test the plugin architecture of the fprime_gds. This file defines a
number of plugins, instantiates a non-singleton plugin instance, and registers the plugins to it. These steps are
provided via the `plugins` fixture to tests.

@author lestarch
"""
import os
import signal
import subprocess
import sys
import time

import pytest

from pathlib import Path
from tempfile import NamedTemporaryFile
from fprime_gds.common.communication.adapters.ip import IpAdapter
from fprime_gds.common.communication.adapters.base import NoneAdapter
from fprime_gds.common.communication.adapters.uart import SerialAdapter
from fprime_gds.common.communication.framing import FramerDeframer, FpFramerDeframer
from fprime_gds.executables.cli import ParserBase, PluginArgumentParser
from fprime_gds.executables.apps import GdsFunction, GdsApp
from fprime_gds.plugin.definitions import gds_plugin_implementation
from fprime_gds.plugin.system import Plugins


class BadSuperClass(object):
    """ A bad framing plugin (inappropriate parent)

    This plugin implementation will fail to register as a "framing" plugin because it does not inherit from
    FramerDeframer.
    """

    @classmethod
    @gds_plugin_implementation
    def register_framing_plugin(cls):
        """ Register a bad plugin """
        return cls


class BadImplementation(FramerDeframer):
    """ A bad framing plugin (inappropriate implementation)

    This plugin implementation will fail to register as a "framing" plugin because it does not implement all abstract
    methods required by the `FramerDeframer` class.
    """

    @classmethod
    @gds_plugin_implementation
    def register_framing_plugin(cls):
        """ Register a bad plugin """
        return cls


class Good(FramerDeframer):
    """ A good framing plugin

    This plugin implementation inherits from the correct class and implements abstract functions. Thus, this plugin will
    register correctly as a "framing" plugin.
    """

    def frame(self, data):
        """ Required function: frame data """
        pass

    def deframe(self, data, no_copy=False):
        """ Required function: deframe data """
        pass

    @classmethod
    @gds_plugin_implementation
    def register_framing_plugin(cls):
        """ Register a good plugin """
        return cls


class GoodWithArgs(Good):
    """ A good framing plugin including arguments

    This plugin implementation is the same "good" implementation with the addition of arguments requested from the
    plugin system. This implementation will take and store these arguments to prove the system.

    Additionally, this plugin implementation defines the `get_name` class method to provide a name for the plugin.
    """
    NAMED_PLUGIN_NAME = "good-with-args"

    def __init__(self, my_fancy_arg, fancy_2):
        """ Fancy argument input processing """
        self.my_fancy_arg = my_fancy_arg
        self.fancy_2 = fancy_2

    @classmethod
    def get_name(cls):
        """ Return the name of this function """
        return cls.NAMED_PLUGIN_NAME

    @classmethod
    def get_arguments(cls):
        """ Return some argument specifications """
        return {
            ("--my-fancy-arg", ): {
                "type": str,
                "help": "Some help string"
            },
            ("--my-fancy-arg-with-dest", ): {
                "dest": "fancy_2",
                "type": int,
                "help": "Some help string"
            },
        }

    @classmethod
    def check_arguments(cls, my_fancy_arg, fancy_2):
        """ Check arguments to raise ValueError """
        if fancy_2 < 0:
            raise ValueError("Must be positive")


class StartFunction(GdsFunction):
    """ A plugin implementation that starts a function

    This plugin implementation creates a start-up function that will run on the invocation of the GDS.
    """
    def __init__(self, start_up_file):
        """ Start-Up function file annotation """
        self.start_up_function_file = start_up_file

    def run(self):
        """ Run function """
        with open(self.start_up_function_file, "a+") as file_handle:
            file_handle.write("ACK\n")

    @classmethod
    def get_name(cls):
        """ Get name """
        return "test-start-function"

    @classmethod
    def get_arguments(cls):
        """ Get arguments """
        return {
            ("--start-up-file", ): {
                "type": str,
                "help": "File to run start-up function on",
                "required": True
            }
        }

    @classmethod
    @gds_plugin_implementation
    def register_gds_function_plugin(cls):
        """ Register a good plugin """
        return cls


class StartApp(GdsApp):
    """ A plugin implementation that starts a app in a separate process

    This plugin implementation creates a start-up application that will run in a separate process
    """
    def __init__(self, start_up_file):
        """ Start-Up function file annotation """
        super().__init__()
        self.start_up_app_file = start_up_file

    def get_process_invocation(self):
        """ Process invocation """
        return [sys.executable, __file__, self.start_up_app_file, "ACK-2"]

    @classmethod
    def get_name(cls):
        """ Get name """
        return "test-start-app"

    @classmethod
    def get_arguments(cls):
        """ Get arguments """
        return {
            ("--start-up-file", ): {
                "type": str,
                "help": "File to run start-up function on",
                "required": True
            }
        }

    @classmethod
    @gds_plugin_implementation
    def register_gds_app_plugin(cls):
        """ Register a good plugin """
        return cls


@pytest.fixture()
def plugins():
    """ Register test plugins as fixture for testing

    Create a `plugins` fixture that is a plugin system created to register specifically the test plugins on top of the
    installed and provided plugins.
    """
    system = Plugins(["communication", "framing"])
    system.register_plugin(BadSuperClass)
    system.register_plugin(BadImplementation)
    system.register_plugin(Good)
    system.register_plugin(GoodWithArgs)
    Plugins._singleton = system
    return system


@pytest.fixture()
def start_up(request):
    """ Register test start up plugins as a fixture for testing

    """
    parent_path = Path(__file__).parent
    extra_plugins, flags = request.param

    command_arguments = ["fprime-gds", "-n", "--dictionary", str( parent_path / "sample" / "dictionary.xml"),
                         "--zmq-transport", "ipc:///tmp/fprime-test-in",  "ipc:///tmp/fprime-test-out",
                         "-g", "none"] + flags

    # Update the environment to side-load python
    environment = os.environ.copy()
    environment[Plugins.PLUGIN_ENVIRONMENT_VARIABLE] = extra_plugins
    environment["PYTHONPATH"] = f"{environment.get('PYTHONPATH', '')}:{parent_path}"
    with NamedTemporaryFile(mode="w+") as temp_file:
        assert "" == temp_file.read(), "Failed to read empty file"
        command_arguments += ["--start-up-file", temp_file.name]
        # Run subprocess for 3 seconds then kill the GDS with 3 seconds to shut down
        process = subprocess.Popen(command_arguments, env=environment)
        time.sleep(3)
        process.send_signal(signal.SIGINT)
        _ = process.communicate(None, 3)
        yield temp_file


def test_base_plugin(plugins):
    """ Tests good framing plugins are returned """
    plugin_options = plugins.get_plugins("framing")
    assert Good in [plugin.plugin_class for plugin in plugin_options], "Good plugin not registered as expected"


def test_framing_builtin_plugins(plugins):
    """ Tests good framing plugins are returned """
    plugin_options = plugins.get_plugins("framing")
    assert FpFramerDeframer in [plugin.plugin_class for plugin in plugin_options], "FpFramerDeframer plugin not registered as expected"


def test_communication_builtin_plugins(plugins):
    """ Tests good framing plugins are returned """
    plugin_options = plugins.get_plugins("communication")
    plugin_classes = [plugin.plugin_class for plugin in plugin_options]
    for expected in [IpAdapter, NoneAdapter, SerialAdapter]:
        assert expected in plugin_classes, f"{expected.__name__} plugin not registered as expected"


def test_plugin_categories(plugins):
    """ Tests plugin categories """
    plugin_categories = plugins.get_categories()
    assert sorted(plugin_categories) == sorted(["communication", "framing"]), "Detected plugin categories incorrect"


def test_plugin_validation(plugins):
    """ Tests good framing plugins are returned """
    plugin_options = plugins.get_plugins("framing")
    plugin_classes = [plugin.plugin_class for plugin in plugin_options]
    assert BadSuperClass not in plugin_classes, "Plugin with bad parent class not excluded as expected"
    assert BadImplementation not in plugin_classes, "Plugin with abstract implementation not excluded as expected"


def test_plugin_arguments(plugins):
    """ Tests that arguments can be parsed and supplied to a plugin """
    a_string = "a_string"
    a_number = "201"
    to_parse = ["--framing", "good-with-args", "--my-fancy-arg", a_string, "--my-fancy-arg-with-dest", a_number]
    args, _ = ParserBase.parse_args([PluginArgumentParser,], arguments=to_parse)
    assert args.framing_selection == GoodWithArgs.NAMED_PLUGIN_NAME, "Improper framing selection"
    assert isinstance(args.framing_selection_instance, GoodWithArgs), "Invalid instance created"
    assert args.framing_selection_instance.my_fancy_arg == a_string, "String argument did not process"
    assert args.framing_selection_instance.fancy_2 == int(a_number), "Integer argument did not process"


def test_plugin_check_arguments(plugins):
    """ Tests that arguments are validated in plugins """
    a_string = "a_string"
    a_number = "-20"
    to_parse = ["--framing", "good-with-args", "--my-fancy-arg", a_string, "--my-fancy-arg-with-dest", a_number]
    with pytest.raises(SystemExit):
        args, _ = ParserBase.parse_args([PluginArgumentParser,], arguments=to_parse)


@pytest.mark.parametrize("start_up", [(f"{__name__}:StartFunction", [])], indirect=True)
def test_start_function(start_up):
    """ Test start-up functions """
    assert "ACK\n" == start_up.read(), "Failed to read expected data"


@pytest.mark.parametrize("start_up", [(f"{__name__}:StartFunction", ["--disable-test-start-function"])], indirect=True)
def test_disabled_start_function(start_up):
    """ Test disabled start-up functions """
    assert "" == start_up.read(), "Failed to read empty file"


@pytest.mark.parametrize("start_up", [(f"{__name__}:StartApp", [])], indirect=True)
def test_start_app(start_up):
    """ Test start-up functions """
    assert "ACK-2\n" == start_up.read(), "Failed to read expected data"


@pytest.mark.parametrize("start_up", [(f"{__name__}:StartApp", ["--disable-test-start-app"])], indirect=True)
def test_disabled_start_app(start_up):
    """ Test disabled start-up functions """
    assert "" == start_up.read(), "Failed to read empty file"


def main():
    """ Run main entry point function

    The main function is run when the file is invoked not as a pytest set of test, but rather when it is run as part of
    the StartApp plugin. This main program writes the second argument to the file represented as the first.
    """
    with open(sys.argv[1], "a+") as file_handle:
        file_handle.write(f"{sys.argv[2]}\n")

if __name__ == "__main__":
    main()



""" fprime_gds.executables.apps: an implementation of start-up apps in fprime

There are twp ways to approach start=up applications in fprime. First, is to implement a run method via a subclass of
`GdsFunction`. This gives the implementor the ability to run anything within the run function that python offers,
however; this comes with complexity of setting up a new thread/process/isolation to ensure that the plugin does not
threaten the fprime-gds core functionality and processes.

The second method is to inherit from `GdsApp` implementing the `get_process_invocation` function to return the necessary
command line that will be spun into its own process.

@author lestarch
"""
import subprocess
from abc import ABC, abstractmethod
from typing import List, Type

from fprime_gds.plugin.definitions import gds_plugin_specification


class GdsBaseFunction(ABC):
    """ Base functionality for pluggable GDS start-up functions

    GDS start-up functionality is pluggable. This class acts as a base for pluggable functionality supplies helpers to
    the various start-up plugins.

    Developers who intend to run in an isolated subprocess are strongly encouraged to use `GdsApp` (see below).
    Developers who need flexibility may use GdsFunction.
    """

    @abstractmethod
    def run(self):
        """ Run the start-up function

        Run the start-up function unconstrained by the limitations of running in a dedicated subprocess.

        """
        raise NotImplementedError()


class GdsFunction(GdsBaseFunction, ABC):
    """ Functionality for pluggable GDS start-up functions

    GDS start-up functionality is pluggable. This class acts as a wide-open implementation of functionality via a single
    `run` callback. Developers have complete control of the start-up functionality. However, this comes at the cost of
    instability in that case of poorly designed functions.

    Developers who intend to run in an isolated subprocess are strongly encouraged to use `GdsApp` (see below).

    Plugin developers are required to implement a single function `run`, which must take care of setting up and running
    the start-up function. Developers **must** handle the isolation of this functionality including spinning off a new
    thread, subprocess, etc. Additionally, the developer must define the `register_gds_function_plugin` class method
    annotated with the @gds_plugin_implementation annotation.

    Standard plug-in functions (get_name, get_arguments) are available should the implementer desire these features.
    Arguments will be supplied to the class's `__init__` function.
    """

    @classmethod
    @gds_plugin_specification
    def register_gds_function_plugin(cls) -> Type["GdsFunction"]:
        """Register gds start-up functionality

        Plugin hook for registering a plugin that supplies start-up functionality. This functionality will run on start-up
        of the GDS network.

        Note: users should return the class, not an instance of the class. Needed arguments for instantiation are
        determined from class methods, solicited via the command line, and provided at construction time to the chosen
        instantiation.

        Returns:
            GDSFunction subclass
        """
        raise NotImplementedError()


class GdsApp(GdsBaseFunction):
    """ GDS start-up process functionality

    A pluggable base class used to start a new process as part of the GDS command line invocation. This allows
    developers to add process-isolated functionality to the GDS network.

    Plugin developers are required to implement the `get_process_invocation` function that returns a list of arguments
    needed to invoke the process via python's `subprocess`. Additionally, the developer must define the
    `register_gds_function_plugin` class method annotated with the @gds_plugin_implementation annotation.

    Standard plug-in functions (get_name, get_arguments) are available should the implementer desire these features.
    Arguments will be supplied to the class's `__init__` function.
    """
    def __init__(self, **arguments):
        """ Construct the communication applications around the arguments

        Command line arguments are passed in to match those returned from the `get_arguments` functions.

        Args:
            arguments: arguments from the command line
        """
        self.process = None
        self.arguments = arguments

    def run(self):
        """ Run the application as an isolated process

        GdsFunction objects require an implementation of the `run` command. This implementation will take the arguments
        provided from `get_process_invocation` function and supplies them as an invocation of the isolated subprocess.
        """
        invocation_arguments = self.get_process_invocation()
        self.process = subprocess.Popen(invocation_arguments)

    def wait(self, timeout=None):
        """ Wait for the app to complete then return the return code

        Waits (blocking) for the process to complete. Then returns the return code of the underlying process. If timeout
        is non-None then the process will be killed after waiting for the timeout and another wait of timeout will be
        allowed for the killed process to exit.

        Return:
            return code of the underlying process
        """
        try:
            _, _ = self.process.wait(timeout=timeout)
        except subprocess.TimeoutExpired:
            self.process.kill()
            _, _ = self.process.wait(timeout=timeout)
        return self.process.returncode

    @abstractmethod
    def get_process_invocation(self) -> List[str]:
        """ Run the start-up function

        Run the start-up function unconstrained by the limitations of running in a dedicated subprocess.

        """
        raise NotImplementedError()

    @classmethod
    @gds_plugin_specification
    def register_gds_app_plugin(cls) -> Type["GdsApp"]:
        """Register a gds start-up application

        Plugin hook for registering a plugin that supplies start-up functionality. This functionality will run on start-up
        of the GDS network isolated into a dedicated process.

        Note: users should return the class, not an instance of the class. Needed arguments for instantiation are
        determined from class methods, solicited via the command line, and provided at construction time to the chosen
        instantiation.

        Returns:
            GdsApp subclass
        """
        raise NotImplementedError()



""" fprime_gds.plugin.system: implementation of plugins

This file contains the implementation and registration of plugins for fprime_gds. Primarily, it defines the Plugins
class that handles plugins. Users can acquire the Plugin singleton with `Plugin.system()`.

This file also imports and registers plugin implementations built-into fprime-gds. These plugins are not registered
using entrypoints.

@author lestarch
"""
import os
import importlib
import inspect
import logging
from typing import Iterable, List, Union

import pluggy

from fprime_gds.plugin.definitions import Plugin, PluginType, PROJECT_NAME

# For automatic validation of plugins, each plugin class type must be imported here
from fprime_gds.executables.apps import GdsFunction, GdsApp
from fprime_gds.common.communication.framing import FramerDeframer, FpFramerDeframer
from fprime_gds.common.communication.adapters.base import BaseAdapter, NoneAdapter
from fprime_gds.common.communication.adapters.ip import IpAdapter

try:
    from fprime_gds.common.communication.adapters.uart import SerialAdapter
except ImportError:
    SerialAdapter = None

# Handy constants
LOGGER = logging.getLogger(__name__)


# Metadata regarding each plugin:
_PLUGIN_METADATA = {
    "framing": {
        "class": FramerDeframer,
        "type": PluginType.SELECTION,
        "built-in": [FpFramerDeframer]
    },
    "communication": {
        "class": BaseAdapter,
        "type": PluginType.SELECTION,
        "built-in": [adapter for adapter in [NoneAdapter, IpAdapter, SerialAdapter] if adapter is not None]
    },
    "gds_function": {
        "class": GdsFunction,
        "type": PluginType.FEATURE,
        "built-in": []
    },
    "gds_app": {
        "class": GdsApp,
        "type": PluginType.FEATURE,
        "built-in": []
    }
}


class PluginException(Exception):
    pass


class InvalidCategoryException(PluginException):
    pass


class Plugins(object):
    """GDS plugin system providing a plugin Singleton for use across the GDS

    GDS plugins are broken into categories (e.g. framing) that represent the key features users can adjust. Each GDS
    application will support and load the plugins for a given category.
    """
    PLUGIN_ENVIRONMENT_VARIABLE = "FPRIME_GDS_EXTRA_PLUGINS"
    _singleton = None

    def __init__(self, categories: Union[None, List] = None):
        """ Initialize the plugin system with specific categories

        Initialize the plugin system with support for the supplied categories. Only plugins for the specified categories
        will be loaded for use. Other plugins will not be available for use.

        Args:
            categories: None for all categories otherwise a list of categories
        """
        categories = self.get_all_categories() if categories is None else categories
        self.categories = categories
        self.manager = pluggy.PluginManager(PROJECT_NAME)

        # Load hook specifications from only the configured categories
        for category in categories:
            self.manager.add_hookspecs(_PLUGIN_METADATA[category]["class"])

        # Load plugins from setuptools entrypoints and the built-in plugins (limited to category)
        self.manager.load_setuptools_entrypoints(PROJECT_NAME)

        # Load plugins from environment variable specified modules
        for token in [token for token in os.environ.get(self.PLUGIN_ENVIRONMENT_VARIABLE, "").split(";") if token]:
            module, class_token = token.split(":")
            try:
                imported_module = importlib.import_module(module)
                module_class = module if class_token == "" else getattr(imported_module, class_token, imported_module)
                self.register_plugin(module_class)
            except ImportError as imp:
                LOGGER.debug("Failed to load %s.%s as plugin", module, class_token)

        # Load built-in plugins
        for category in categories:
            for built_in in _PLUGIN_METADATA[category]["built-in"]:
                self.register_plugin(built_in)

    def get_plugins(self, category) -> Iterable:
        """Get available plugins for the given category

        Gets all plugin implementors of "category" by looking for register_<category>_plugin implementors. If such a
        function does not exist then this results in an exception.

        Args:
            category: category of the plugin requested

        Return:
            validated list of plugin implementor classes
        """
        try:
            plugin_classes = getattr(self.manager.hook, f"register_{category}_plugin")()
        except KeyError as error:
            raise InvalidCategoryException(f"Invalid plugin category: {error}")

        return [
            Plugin(category, self.get_category_plugin_type(category), plugin_class)
            for plugin_class in plugin_classes
            if self.validate_selection(category, plugin_class)
        ]

    def register_plugin(self, module_or_class):
        """Register a plugin directly

        Allows local registration of plugin implementations that are shipped as part of the GDS package.

        Args:
            module_or_class: module or class that has plugin implementations
        """
        self.manager.register(module_or_class)

    def get_categories(self):
        """ Get plugin categories """
        return self.categories

    @staticmethod
    def get_all_categories():
        """ Get all plugin categories """
        return _PLUGIN_METADATA.keys()

    @staticmethod
    def get_plugin_metadata(category):
        """ Get the plugin metadata for a given plugin category """
        return _PLUGIN_METADATA[category]

    @classmethod
    def get_category_plugin_type(cls, category):
        """ Get the plugin type given the category """
        return cls.get_plugin_metadata(category)["type"]

    @classmethod
    def get_category_specification_class(cls, category):
        """ Get the plugin class given the category """
        return cls.get_plugin_metadata(category)["class"]

    @classmethod
    def validate_selection(cls, category, result):
        """Validate the result of plugin hook

        Validates the result of a plugin hook call to ensure the result meets the expected properties for plugins of the
        given category. Primarily this ensures that this plugin returns a concrete subclass of the expected type.

        Args:
            category: category of plugin used
            result: result from the plugin hook call
        Return:
            True when the plugin passes validation, False otherwise
        """
        # Typing library not intended for introspection at runtime, thus we maintain a map of plugin specification
        # functions to the types expected as a return value. When this is not found, plugins may continue without
        # automatic validation.
        try:
            expected_class = cls.get_category_specification_class(category)
            # Validate the result
            if not issubclass(result, expected_class):
                LOGGER.warning(
                    f"{result.__name__} is not a subclass of {expected_class.__name__}. Not registering."
                )
                return False
            elif inspect.isabstract(result):
                LOGGER.warning(
                    f"{result.__name__} is an abstract class. Not registering."
                )
                return False
        except KeyError:
            LOGGER.warning(
                f"Plugin not registered for validation. Continuing without validation."
            )
        return True

    @classmethod
    def system(cls, categories: Union[None, List] = None) -> "Plugins":
        """ Get plugin system singleton

        Constructs the plugin system singleton (when it has yet to be constructed) then returns the singleton. The
        singleton will support specific categories and further requests for a singleton will cause an assertion error
        unless the categories match or is None.

        Args:
            categories: a list of categories to support or None to use the existing categories

        Returns:
            plugin system
        """
        # Singleton undefined, construct it
        if cls._singleton is None:
            cls._singleton = cls(cls.get_all_categories() if categories is None else categories)
        # Ensure categories was unspecified or matches the singleton
        assert categories is None or cls._singleton.categories == categories, "Inconsistent plugin categories"
        return cls._singleton




""" fprime_gds.plugin.definitions: definitions of plugin specifications and decorators

In order to define a plugin, an implementation decorator is used. Users can import `gds_plugin_implementation` from this
file to decorate functions that implement plugins.

This file also defines helper classes to support the plugin system.

@author lestarch
"""
import pluggy
from enum import Enum, auto
from typing import Any, Dict, Tuple, Type

PROJECT_NAME = "fprime_gds"

gds_plugin_specification = pluggy.HookspecMarker(PROJECT_NAME)
gds_plugin_implementation = pluggy.HookimplMarker(PROJECT_NAME)


class PluginType(Enum):
    """ Enumeration of plugin types"""
    ALL = auto()
    """ Plugin selection including all types of plugins """

    SELECTION = auto()
    """ Plugin that provides a selection between implementations """

    FEATURE = auto()
    """ Plugin that provides a feature """


class Plugin(object):
    """ Plugin wrapper object """

    def __init__(self, category: str, plugin_type: PluginType, plugin_class: Type[Any]):
        """ Initialize the plugin

        Args:
            category: category of the plugin (i.e. register_<category>_function)
            plugin_type: type of plugin
            plugin_class: implementation class of the plugin
        """
        self.category = category
        self.type = plugin_type
        self.plugin_class = plugin_class

    def get_name(self):
        """ Get the name of the plugin

        Plugin names are derived from the `get_name` class method of the plugin's implementation class. When not defined
        that name is derived from the plugin's implementation class __name__ property instead.

        Returns:
            name of plugin
        """
        return (
            self.plugin_class.get_name() if hasattr(self.plugin_class, "get_name")
            else self.plugin_class.__name__
        )

    def get_arguments(self) -> Dict[Tuple[str, ...], Dict[str, Any]]:
        """ Get arguments needed by plugin

        Plugin argument are derived from the `get_arguments` class method of the plugin's implementation class. When not
        defined an empty dictionary is returned.

        Returns:
            argument specification for plugin
        """
        return self.plugin_class.get_arguments() if hasattr(self.plugin_class, "get_arguments") else {}




"""
cli.py:

This file sets up the command line interface and argument parsing that is done to support the F prime executable tools
layer. It is designed to allow users to import standard sets of arguments that applied to the various aspects of the
code that they are importing.

@author mstarch
"""

import argparse
import datetime
import errno
import getpass
import itertools
import os
import platform
import re
import sys
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any, Dict, List, Tuple

# Required to set the checksum as a module variable
import fprime_gds.common.communication.checksum
import fprime_gds.common.logger
from fprime_gds.common.communication.adapters.ip import check_port
from fprime_gds.common.pipeline.standard import StandardPipeline
from fprime_gds.common.transport import ThreadedTCPSocketClient
from fprime_gds.common.utils.config_manager import ConfigManager
from fprime_gds.executables.utils import find_app, find_dict, get_artifacts_root
from fprime_gds.plugin.definitions import PluginType
from fprime_gds.plugin.system import Plugins
from fprime_gds.common.zmq_transport import ZmqClient


GUIS = ["none", "html"]


class ParserBase(ABC):
    """Base parser for handling fprime command lines

    Parsers must define several functions. They must define "get_parser", which will produce a parser to parse the
    arguments, and an optional "handle_arguments" function to do any necessary processing of the arguments. Note: when
    handling arguments.
    """

    DESCRIPTION = None

    @property
    def description(self):
        """Return parser description"""
        return self.DESCRIPTION if self.DESCRIPTION else "Unknown command line parser"

    @abstractmethod
    def get_arguments(self) -> Dict[Tuple[str, ...], Dict[str, Any]]:
        """Return argument list handled by this parser

        Produce the arguments that can be processed by multiple parsers. i.e. argparse, and pytest parsers are the
        intended consumers. Returns a tuple of dictionary of flag tuples (--flag, -f) to keyword arguments to pass to
        argparse and list of arguments calculated by the parser (generated).

        Returns:
            tuple of dictionary of flag tuple to keyword arguments, list of generated fields
        """

    def get_parser(self) -> argparse.ArgumentParser:
        """Return an argument parser to parse arguments here-in

        Produce a parser that will handle the given arguments. These parsers can be combined for a CLI for a tool by
        assembling them as parent processors to a parser for the given tool.

        Return:
            argparse parser for supplied arguments
        """
        parser = argparse.ArgumentParser(
            description=self.description,
            add_help=True,
            formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        )
        self.fill_parser(parser)
        return parser

    def fill_parser(self, parser):
        """ Fill supplied parser with arguments

        Fills the supplied parser with the arguments returned via the `get_arguments` method invocation. This
        implementation add the arguments directly to the parser.

        Args:
            parser: parser to fill with arguments

        """
        for flags, keywords in self.get_arguments().items():
            try:
                parser.add_argument(*flags, **keywords)
            except argparse.ArgumentError:
                # flag has already been added, pass
                pass

    def reproduce_cli_args(self, args_ns):
        """Reproduce the list of arguments needed on the command line"""

        def flag_member(flags, argparse_inputs) -> Tuple[str, str]:
            """Get the best CLI flag and namespace member"""
            best_flag = (
                [flag for flag in flags if flag.startswith("--")] + list(flags)
            )[0]
            member = argparse_inputs.get(
                "dest", re.sub(r"^-+", "", best_flag).replace("-", "_")
            )
            return best_flag, member

        def cli_arguments(flags, argparse_inputs) -> List[str]:
            """Get CLI argument list fro argument entry"""
            best_flag, member = flag_member(flags, argparse_inputs)
            value = getattr(args_ns, member, None)

            action = argparse_inputs.get("action", "store")
            assert action in [
                "store",
                "store_true",
                "store_false",
            ], f"{action} not supported by reproduce_cli_args"

            # Handle arguments
            if (action == "store_true" and value) or (
                action == "store_false" and not value
            ):
                return [best_flag]
            elif action != "store" or value is None:
                return []
            return [best_flag] + (
                [str(item) for item in value]
                if isinstance(value, list)
                else [str(value)]
            )

        cli_pairs = [
            cli_arguments(flags, argparse_ins)
            for flags, argparse_ins in self.get_arguments().items()
        ]
        return list(itertools.chain.from_iterable(cli_pairs))

    @abstractmethod
    def handle_arguments(self, args, **kwargs):
        """Post-process the parser's arguments

        Handle arguments from the given parser. The expectation is that the "args" namespace is taken in, processed, and
        a new namespace object is returned with the processed variants of the arguments.

        Args:
            args: arguments namespace of processed arguments
        Returns: namespace with processed results of arguments.
        """

    @staticmethod
    def parse_args(
        parser_classes,
        description="No tool description provided",
        arguments=None,
        **kwargs,
    ):
        """Parse and post-process arguments

        Create a parser for the given application using the description provided. This will then add all specified
        ParserBase subclasses' get_parser output as parent parses for the created parser. Then all of the handle
        arguments methods will be called, and the final namespace will be returned.

        Args:
            parser_classes: a list of ParserBase subclasses that will be used to
            description: description passed ot the argument parser
            arguments: arguments to process, None to use command line input
        Returns: namespace with all parsed arguments from all provided ParserBase subclasses
        """
        composition = CompositeParser(parser_classes, description)
        parser = composition.get_parser()
        try:
            args_ns = parser.parse_args(arguments)
            args_ns = composition.handle_arguments(args_ns, **kwargs)
        except ValueError as ver:
            print(f"[ERROR] Failed to parse arguments: {ver}", file=sys.stderr)
            parser.print_help()
            sys.exit(-1)
        except Exception as exc:
            print(f"[ERROR] {exc}", file=sys.stderr)
            sys.exit(-1)
        return args_ns, parser

    @staticmethod
    def find_in(token, deploy, is_file=True):
        """
        Find token in deploy directory by walking the directory looking for reg-ex. This effectively finds a file in a
        subtree and provides the path to it. Returns None when not found

        :param token: token to search for in the directory structure
        :param deploy: directory to start with
        :param is_file: true if looking for file, otherwise false
        :return: full path to token in tree
        """
        for dirpath, dirs, files in os.walk(deploy):
            for check in files if is_file else dirs:
                if re.match(f"^{str(token)}$", check):
                    return os.path.join(dirpath, check)
        return None


class DetectionParser(ParserBase):
    """Parser that detects items from a root/directory or deployment"""

    def get_arguments(self) -> Dict[Tuple[str, ...], Dict[str, Any]]:
        """Arguments needed for root processing"""
        return {
            ("-d", "--deployment"): {
                "dest": "deployment",
                "action": "store",
                "required": False,
                "type": str,
                "help": "Deployment installation/build output directory. [default: install_dest field in settings.ini]",
            }
        }

    def handle_arguments(self, args, **kwargs):
        """Handle the root, detecting it if necessary"""
        if args.deployment:
            args.deployment = Path(args.deployment)
            return args
        detected_toolchain = get_artifacts_root() / platform.system()
        if not detected_toolchain.exists():
            msg = f"{detected_toolchain} does not exist. Make sure to build."
            raise Exception(msg)
        likely_deployment = detected_toolchain / Path.cwd().name
        # Check if the deployment exists
        if likely_deployment.exists():
            args.deployment = likely_deployment
            return args
        child_directories = [
            child for child in detected_toolchain.iterdir() if child.is_dir()
        ]
        if not child_directories:
            msg = f"No deployments found in {detected_toolchain}. Specify deployment with: --deployment"
            raise Exception(msg)
        # Works for the old structure where the bin, lib, and dict directories live immediately under the platform
        elif len(child_directories) == 3 and set(
            [path.name for path in child_directories]
        ) == {"bin", "lib", "dict"}:
            args.deployment = detected_toolchain
            return args
        elif len(child_directories) > 1:
            msg = f"Multiple deployments found in {detected_toolchain}. Choose using: --deployment"
            raise Exception(msg)
        args.deployment = child_directories[0]
        return args


class PluginArgumentParser(ParserBase):
    """Parser for arguments coming from plugins"""

    DESCRIPTION = "Plugin options"
    FPRIME_CHOICES = {
        "framing": "fprime",
        "communication": "ip",
    }

    def __init__(self):
        """Initialize the plugin information for this parser"""
        self._plugin_map = {
            category: Plugins.system().get_plugins(category)
            for category in Plugins.system().get_categories()
        }

    @staticmethod
    def safe_add_argument(parser, *flags, **keywords):
        """ Add an argument allowing duplicates

        Add arguments to the parser (passes through *flags and **keywords) to the supplied parser. This method traps
        errors to prevent duplicates.

        Args:
            parser: parser or argument group to add arguments to
            *flags: positional arguments passed to `add_argument`
            **keywords: key word arguments passed to `add_argument`
        """
        try:
            parser.add_argument(*flags, **keywords)
        except argparse.ArgumentError:
            # flag has already been added, pass
            pass

    def fill_parser(self, parser):
        """ File supplied parser with grouped arguments

        Fill the supplied parser with arguments from the `get_arguments` method invocation. This implementation groups
        arguments based on the constituent that sources the argument.

        Args:
            parser: parser to fill
        """
        for category, plugins in self._plugin_map.items():
            argument_group = parser.add_argument_group(title=f"{category.title()} Plugin Options")
            for flags, keywords in self.get_category_arguments(category).items():
                self.safe_add_argument(argument_group, *flags, **keywords)

            for plugin in plugins:
                argument_group = parser.add_argument_group(title=f"{category.title()} Plugin '{plugin.get_name()}' Options")
                if plugin.type == PluginType.FEATURE:
                    self.safe_add_argument(argument_group,
                                           f"--disable-{plugin.get_name()}",
                                           action="store_true",
                                           default=False,
                                           help=f"Disable the {category} plugin '{plugin.get_name()}'")
                for flags, keywords in plugin.get_arguments().items():
                    self.safe_add_argument(argument_group, *flags, **keywords)

    def get_category_arguments(self, category):
        """ Get arguments for a plugin category """
        arguments: Dict[Tuple[str, ...], Dict[str, Any]] = {}
        plugins = self._plugin_map[category]
        # Add category options: SELECTION plugins add a selection flag
        plugin_type = Plugins.get_category_plugin_type(category)
        if plugin_type == PluginType.SELECTION:
            arguments.update(
                {
                    (f"--{category}-selection",): {
                        "choices": [choice.get_name() for choice in plugins],
                        "help": f"Select {category} implementer.",
                        "default": self.FPRIME_CHOICES.get(
                            category, list(plugins)[0].get_name()
                        ),
                    }
                }
            )
        return arguments

    def get_arguments(self) -> Dict[Tuple[str, ...], Dict[str, Any]]:
        """Return arguments to used in plugins"""
        arguments: Dict[Tuple[str, ...], Dict[str, Any]] = {}
        for category, plugins in self._plugin_map.items():
            arguments.update(self.get_category_arguments(category))
            for plugin in plugins:
                # Add disable flags for feature type plugins
                if plugin.type == PluginType.FEATURE:
                    arguments.update({
                        (f"--disable-{plugin.get_name()}", ): {
                            "action": "store_true",
                            "default": False,
                            "help": f"Disable the {category} plugin '{plugin.get_name()}'"
                        }
                    })
                arguments.update(plugin.get_arguments())
        return arguments

    def handle_arguments(self, args, **kwargs):
        """Handles the arguments"""
        for category, plugins in self._plugin_map.items():
            plugin_type = Plugins.get_category_plugin_type(category)

            # Selection plugins choose one plugin and instantiate it
            if plugin_type == PluginType.SELECTION:
                selection_string = getattr(args, f"{category}_selection")
                matching_plugins = [plugin for plugin in plugins if plugin.get_name() == selection_string]
                assert len(matching_plugins) == 1, "Plugin selection system failed"
                selection_class = matching_plugins[0].plugin_class
                filled_arguments = self.extract_plugin_arguments(args, selection_class)
                selection_instance = selection_class(**filled_arguments)
                setattr(args, f"{category}_selection_instance", selection_instance)
            # Feature plugins instantiate all enabled plugins
            elif plugin_type == PluginType.FEATURE:
                enabled_plugins = [
                    plugin for plugin in plugins
                    if not getattr(args, f"disable_{plugin.get_name().replace('-', '_')}", False)
                ]
                plugin_instantiations = [
                    plugin.plugin_class(**self.extract_plugin_arguments(args, plugin))
                    for plugin in enabled_plugins
                ]
                setattr(args, f"{category}_enabled_instances", plugin_instantiations)
        return args

    @staticmethod
    def extract_plugin_arguments(args, plugin) -> Dict[str, Any]:
        """Extract plugin argument values from the args namespace into a map

        Plugin arguments will be supplied to the `__init__` function of the plugin via a keyword argument dictionary.
        This function maps from the argument namespace from parsing back into that dictionary.

        Args:
            args: argument namespace from argparse
            plugin: plugin to extract arguments for
        Return:
            filled arguments dictionary
        """
        expected_args = plugin.get_arguments()
        argument_destinations = [
            (
                value["dest"]
                if "dest" in value
                else key[0].replace("--", "").replace("-", "_")
            )
            for key, value in expected_args.items()
        ]
        filled_arguments = {
            destination: getattr(args, destination)
            for destination in argument_destinations
        }
        # Check arguments or yield a Value error
        if hasattr(plugin, "check_arguments"):
            plugin.check_arguments(**filled_arguments)
        return filled_arguments


class CompositeParser(ParserBase):
    """Composite parser handles parsing as a composition of multiple other parsers"""

    def __init__(self, constituents, description=None):
        """Construct this parser by instantiating the sub-parsers"""
        self.given = description
        constructed = [constituent() for constituent in constituents]
        flattened = [
            item.constituents if isinstance(item, CompositeParser) else [item]
            for item in constructed
        ]
        self.constituent_parsers = {*itertools.chain.from_iterable(flattened)}

    def fill_parser(self, parser):
        """ File supplied parser with grouped arguments

        Fill the supplied parser with arguments from the `get_arguments` method invocation. This implementation groups
        arguments based on the constituent that sources the argument.

        Args:
            parser: parser to fill
        """
        for constituent in sorted(self.constituents, key=lambda x: x.description):
            if isinstance(constituent, (PluginArgumentParser, CompositeParser)):
                constituent.fill_parser(parser)
            else:
                argument_group = parser.add_argument_group(title=constituent.description)
                constituent.fill_parser(argument_group)

    @property
    def constituents(self):
        """Get constituent"""
        return self.constituent_parsers

    @property
    def description(self):
        """Return parser description"""
        return (
            self.given
            if self.given
            else ",".join(item.description for item in self.constituents)
        )

    def get_arguments(self) -> Dict[Tuple[str, ...], Dict[str, Any]]:
        """Get the argument from all constituents"""
        arguments = {}
        for constituent in self.constituents:
            arguments.update(constituent.get_arguments())
        return arguments

    def handle_arguments(self, args, **kwargs):
        """Process all constituent arguments"""
        for constituent in self.constituents:
            args = constituent.handle_arguments(args, **kwargs)
        return args


class CommExtraParser(ParserBase):
    """Parses extra communication arguments"""

    DESCRIPTION = "Communications options"

    def get_arguments(self) -> Dict[Tuple[str, ...], Dict[str, Any]]:
        """Get arguments for the comm-layer parser"""
        com_arguments = {
            ("--output-unframed-data",): {
                "dest": "output_unframed_data",
                "action": "store",
                "nargs": "?",
                "help": "Log unframed data to supplied file relative to log directory. Use '-' for standard out.",
                "default": None,
                "const": "unframed.log",
                "required": False,
            },
        }
        return com_arguments

    def handle_arguments(self, args, **kwargs):
        return args


class LogDeployParser(ParserBase):
    """
    A parser that handles log files by reading in a '--logs' directory or a '--deploy' directory to put the logs into
    as a default. This is useful as a parsing fragment for any application that produces log files and needs these logs
    to end up in the proper place.
    """

    DESCRIPTION = "Logging options"

    def get_arguments(self) -> Dict[Tuple[str, ...], Dict[str, Any]]:
        """Return arguments to parse logging options"""
        return {
            ("-l", "--logs"): {
                "dest": "logs",
                "action": "store",
                "default": os.path.join(os.getcwd(), "logs"),
                "type": str,
                "help": "Logging directory. Created if non-existent. [default: %(default)s]",
            },
            ("--log-directly",): {
                "dest": "log_directly",
                "action": "store_true",
                "default": False,
                "help": "Logging directory is used directly, no extra dated directories created.",
            },
            ("--log-to-stdout",): {
                "action": "store_true",
                "default": False,
                "help": "Log to standard out along with log output files",
            },
        }

    def handle_arguments(self, args, **kwargs):
        """
        Read the arguments specified in this parser and validate the expected inputs.

        :param args: parsed arguments as namespace
        :return: args namespace
        """
        # Get logging dir
        if not args.log_directly:
            args.logs = os.path.abspath(
                os.path.join(
                    args.logs, datetime.datetime.now().strftime("%Y_%m_%d-%H_%M_%S")
                )
            )
            # A dated directory has been set, all log handling must now be direct
            args.log_directly = True

        # Make sure directory exists
        try:
            os.makedirs(args.logs, exist_ok=True)
        except OSError as osexc:
            if osexc.errno != errno.EEXIST:
                raise
        # Setup the basic python logging
        fprime_gds.common.logger.configure_py_log(
            args.logs, mirror_to_stdout=args.log_to_stdout
        )
        return args


class MiddleWareParser(ParserBase):
    """
    Middleware (ThreadedTcpServer, ZMQ) interface that looks for an address and a port. The argument handling will
    attempt to connect to the socket to ensure that it is a valid address/port and report any errors. This is then
    immediately closes the port after use. There is a minor race-condition between this check and the actual usage,
    however; it should be close enough.
    """

    DESCRIPTION = "Middleware options"

    def get_arguments(self) -> Dict[Tuple[str, ...], Dict[str, Any]]:
        """Return arguments necessary to run a and connect to the GDS middleware"""
        zmq_arguments = {
            ("--no-zmq",): {
                "dest": "zmq",
                "action": "store_false",
                "help": "Disable ZMQ transportation layer, falling back to TCP socket server.",
                "default": True,
            },
            ("--zmq-transport",): {
                "dest": "zmq_transport",
                "nargs": 2,
                "help": "Pair of URls used with --zmq to setup ZeroMQ transportation [default: %(default)s]",
                "default": [
                    "ipc:///tmp/fprime-server-in",
                    "ipc:///tmp/fprime-server-out",
                ],
                "metavar": ("serverInUrl", "serverOutUrl"),
            },
        }
        tts_arguments = {
            ("--tts-port",): {
                "dest": "tts_port",
                "action": "store",
                "type": int,
                "help": "Set the threaded TCP socket server port when ZMQ is not used [default: %(default)s]",
                "default": 50050,
            },
            ("--tts-addr",): {
                "dest": "tts_addr",
                "action": "store",
                "type": str,
                "help": "Set the threaded TCP socket server address when ZMQ is not used [default: %(default)s]",
                "default": "0.0.0.0",
            },
        }
        return {**zmq_arguments, **tts_arguments}

    def handle_arguments(self, args, **kwargs):
        """
        Checks to ensure that the specified port and address is available before connecting. This prevents user from
        attempting to run on a port that is unavailable.

        :param args: parsed argument namespace
        :return: args namespace
        """
        is_client = kwargs.get("client", False)
        tts_connection_address = (
            args.tts_addr.replace("0.0.0.0", "127.0.0.1")
            if is_client
            else args.tts_addr
        )

        args.connection_uri = f"tcp://{tts_connection_address}:{args.tts_port}"
        args.connection_transport = ThreadedTCPSocketClient
        if args.zmq:
            args.connection_uri = args.zmq_transport
            args.connection_transport = ZmqClient
        elif not is_client:
            check_port(args.tts_addr, args.tts_port)
        return args


class DictionaryParser(DetectionParser):
    """Parser for deployments"""

    DESCRIPTION = "Dictionary options"

    def get_arguments(self) -> Dict[Tuple[str, ...], Dict[str, Any]]:
        """Arguments to handle deployments"""
        return {
            **super().get_arguments(),
            **{
                ("--dictionary",): {
                    "dest": "dictionary",
                    "action": "store",
                    "default": None,
                    "required": False,
                    "type": str,
                    "help": "Path to dictionary. Overrides automatic dictionary detection.",
                },
                ("--packet-spec",): {
                    "dest": "packet_spec",
                    "action": "store",
                    "default": None,
                    "required": False,
                    "type": str,
                    "help": "Path to packet specification.",
                },
            },
        }

    def handle_arguments(self, args, **kwargs):
        """Handle arguments as parsed"""
        # Find dictionary setting via "dictionary" argument or the "deploy" argument
        if args.dictionary is not None and not os.path.exists(args.dictionary):
            msg = f"Dictionary file {args.dictionary} does not exist"
            raise ValueError(msg)
        elif args.dictionary is None:
            args = super().handle_arguments(args, **kwargs)
            args.dictionary = find_dict(args.deployment)
        return args


class FileHandlingParser(ParserBase):
    """Parser for deployments"""

    DESCRIPTION = "File handling options"

    def get_arguments(self) -> Dict[Tuple[str, ...], Dict[str, Any]]:
        """Arguments to handle deployments"""

        username = getpass.getuser()

        return {
            ("--file-storage-directory",): {
                "dest": "files_storage_directory",
                "action": "store",
                "default": "/tmp/" + username,
                "required": False,
                "type": str,
                "help": "Directory to store uplink and downlink files. Default: %(default)s",
            },
            ("--remote-sequence-directory",): {
                "dest": "remote_sequence_directory",
                "action": "store",
                "default": "/seq",
                "required": False,
                "type": str,
                "help": "Directory to save command sequence binaries, on the remote FSW. Default: %(default)s",
            },
        }

    def handle_arguments(self, args, **kwargs):
        """Handle arguments as parsed"""
        try:
            Path(args.files_storage_directory).mkdir(parents=True, exist_ok=True)
        except PermissionError:
            raise PermissionError(
                f"{args.files_storage_directory} is not writable. Fix permissions or change storage directory with --file-storage-directory."
            )
        return args


class StandardPipelineParser(CompositeParser):
    """Standard pipeline argument parser: combination of MiddleWare and"""

    CONSTITUENTS = [
        DictionaryParser,
        FileHandlingParser,
        MiddleWareParser,
        LogDeployParser,
    ]

    def __init__(self):
        """Initialization"""
        super().__init__(
            constituents=self.CONSTITUENTS, description="Standard pipeline setup"
        )

    @staticmethod
    def pipeline_factory(args_ns, pipeline=None) -> StandardPipeline:
        """A factory of the standard pipeline given the handled arguments"""
        pipeline_arguments = {
            "config": ConfigManager(),
            "dictionary": args_ns.dictionary,
            "file_store": args_ns.files_storage_directory,
            "packet_spec": args_ns.packet_spec,
            "logging_prefix": args_ns.logs,
        }
        pipeline = pipeline if pipeline else StandardPipeline()
        pipeline.transport_implementation = args_ns.connection_transport
        try:
            pipeline.setup(**pipeline_arguments)
            pipeline.connect(args_ns.connection_uri)
        except Exception:
            # In all error cases, pipeline should be shutdown before continuing with exception handling
            try:
                pipeline.disconnect()
            finally:
                raise
        return pipeline


class CommParser(CompositeParser):
    """Comm Executable Parser"""

    CONSTITUENTS = [
        CommExtraParser,
        MiddleWareParser,
        LogDeployParser,
        PluginArgumentParser,
    ]

    def __init__(self):
        """Initialization"""
        super().__init__(
            constituents=self.CONSTITUENTS,
            description="Communications bridge application",
        )


class GdsParser(ParserBase):
    """
    Provides a parser for the following arguments:

    - dictionary: path to dictionary, either a folder for py_dicts, or a file for XML dicts
    - logs: path to logging path
    - config: configuration for GDS.

    Note: deployment can help in setting both dictionary and logs, but isn't strictly required.
    """

    DESCRIPTION = "GUI options"

    def get_arguments(self) -> Dict[Tuple[str, ...], Dict[str, Any]]:
        """Return arguments necessary to run a binary deployment via the GDS"""
        return {
            ("-g", "--gui"): {
                "choices": GUIS,
                "dest": "gui",
                "type": str,
                "help": "Set the desired GUI system for running the deployment. [default: %(default)s]",
                "default": "html",
            },
            ("--gui-addr",): {
                "dest": "gui_addr",
                "action": "store",
                "default": "127.0.0.1",
                "required": False,
                "type": str,
                "help": "Set the GUI server address [default: %(default)s]",
            },
            ("--gui-port",): {
                "dest": "gui_port",
                "action": "store",
                "default": "5000",
                "required": False,
                "type": str,
                "help": "Set the GUI server address [default: %(default)s]",
            },
        }

    def handle_arguments(self, args, **kwargs):
        """
        Takes the arguments from the parser, and processes them into the needed map of key to dictionaries for the
        program. This will throw if there is an error.

        :param args: parsed args into a namespace
        :return: args namespace
        """
        return args


class BinaryDeployment(DetectionParser):
    """
    Parsing subclass used to read the arguments of the binary application. This derives functionality from a comm parser
    and represents the flight-side of the equation.
    """

    DESCRIPTION = "FPrime binary options"

    def get_arguments(self) -> Dict[Tuple[str, ...], Dict[str, Any]]:
        """Return arguments necessary to run a binary deployment via the GDS"""
        return {
            **super().get_arguments(),
            **{
                ("-n", "--no-app"): {
                    "dest": "noapp",
                    "action": "store_true",
                    "default": False,
                    "help": "Do not run deployment binary. Overrides --app.",
                },
                ("--app",): {
                    "dest": "app",
                    "action": "store",
                    "required": False,
                    "type": str,
                    "help": "Path to app to run. Overrides automatic app detection.",
                },
            },
        }

    def handle_arguments(self, args, **kwargs):
        """
        Takes the arguments from the parser, and processes them into the needed map of key to dictionaries for the
        program. This will throw if there is an error.

        :param args: parsed arguments in namespace
        :return: args namespaces
        """
        # No app, stop processing now
        if args.noapp:
            return args
        args = super().handle_arguments(args, **kwargs)
        args.app = Path(args.app) if args.app else Path(find_app(args.deployment))
        if not args.app.is_file():
            msg = f"F prime binary '{args.app}' does not exist or is not a file"
            raise ValueError(msg)
        return args


class SearchArgumentsParser(ParserBase):
    """Parser for search arguments"""

    DESCRIPTION = (
        "Searching and filtering options"
    )

    def __init__(self, command_name: str) -> None:
        self.command_name = command_name

    def get_arguments(self) -> Dict[Tuple[str, ...], Dict[str, Any]]:
        """Return arguments necessary to search through channels/events/commands"""
        return {
            ("--list",): {
                "dest": "is_printing_list",
                "action": "store_true",
                "help": f"list all possible {self.command_name[:-1]} types the current F Prime instance could produce, based on the {self.command_name} dictionary, sorted by {self.command_name[:-1]} type ID",
            },
            ("-i", "--ids"): {
                "dest": "ids",
                "action": "store",
                "required": False,
                "type": int,
                "nargs": "+",
                "help": f"only show {self.command_name} matching the given type ID(s) 'ID'; can provide multiple IDs to show all given types",
                "metavar": "ID",
            },
            ("-c", "--components"): {
                "dest": "components",
                "nargs": "+",
                "required": False,
                "type": str,
                "help": f"only show {self.command_name} from the given component name 'COMP'; can provide multiple components to show {self.command_name} from all components given",
                "metavar": "COMP",
            },
            ("-s", "--search"): {
                "dest": "search",
                "required": False,
                "type": str,
                "help": f'only show {self.command_name} whose name or output string exactly matches or contains the entire given string "STRING"',
            },
        }

    def handle_arguments(self, args, **kwargs):
        return args


class RetrievalArgumentsParser(ParserBase):
    """Parser for retrieval arguments"""

    DESCRIPTION = "Data retrieval options"

    def __init__(self, command_name: str) -> None:
        self.command_name = command_name

    def get_arguments(self) -> Dict[Tuple[str, ...], Dict[str, Any]]:
        """Return arguments to retrieve channels/events/commands in specific ways"""
        return {
            ("-t", "--timeout"): {
                "dest": "timeout",
                "action": "store",
                "required": False,
                "type": float,
                "help": f"wait at most SECONDS seconds for a single new {self.command_name[:-1]}, then exit (defaults to listening until the user exits via CTRL+C, and logging all {self.command_name})",
                "metavar": "SECONDS",
                "default": 0.0,
            },
            ("-j", "--json"): {
                "dest": "json",
                "action": "store_true",
                "required": False,
                "help": "returns response in JSON format",
            },
        }

    def handle_arguments(self, args, **kwargs):
        return args
